@@ -498,7 +498,11 @@ class WidgetTest(GuiTest):
         if widget is None:
             widget = self.widget
 
-        widget.unconditional_commit()
+        if hasattr(widget.commit, "now"):
+            widget.commit.now()
+        else:
+            # Support for deprecated, non-decorated commit
+            widget.unconditional_commit()
         self.wait_until_finished(widget=widget, timeout=wait)
 
     def get_output(self, output, widget=None, wait=DEFAULT_TIMEOUT):
@@ -742,7 +746,7 @@ def open_widget_classes():
 
 
 @contextmanager
-def override_default_settings(widget, defaults=None, handler=None):
+def override_default_settings(widget, defaults=None, context_defaults=[], handler=None):
     if defaults is None:
         defaults = {}
 
@@ -754,6 +758,7 @@ def override_default_settings(widget, defaults=None, handler=None):
     os.makedirs(os.path.dirname(filename), exist_ok=True)
     with open(filename, "wb") as f:
         pickle.dump(defaults, f)
+        pickle.dump(context_defaults, f)
 
     yield
 
