@@ -14,6 +14,7 @@ from orangecanvas.registry.description import (
     InputSignal, OutputSignal, Single, Multiple, Default, NonDefault,
     Explicit, Dynamic
 )
+from orangewidget.workflow.utils import WeakKeyDefaultDict
 
 
 # increasing counter for ensuring the order of Input/Output definitions
@@ -76,7 +77,7 @@ def can_summarize(type_, name, explicit):
         if summarize.dispatch(a_type) is base_summarize:
             warnings.warn(
                 f"register 'summarize' function for type {a_type.__name__}. "
-                + instr, UserWarning)
+                + instr, UserWarning, stacklevel=4)
             return False
     return True
 
@@ -187,13 +188,16 @@ class Input(InputSignal, _Signal):
         """
         if self.flags & Multiple:
             def summarize_wrapper(widget, value, id=None):
-                widget.set_partial_input_summary(
-                    self.name, summarize(value), id=id)
+                # If this method is overridden, don't summarize
+                if summarize_wrapper is getattr(type(widget), method.__name__):
+                    widget.set_partial_input_summary(
+                        self.name, summarize(value), id=id)
                 method(widget, value, id)
         else:
             def summarize_wrapper(widget, value):
-                widget.set_partial_input_summary(
-                    self.name, summarize(value))
+                if summarize_wrapper is getattr(type(widget), method.__name__):
+                    widget.set_partial_input_summary(
+                        self.name, summarize(value))
                 method(widget, value)
 
         # Re-binding with the same name can happen in derived classes
@@ -256,10 +260,20 @@ class MultiInput(Input):
         self.filter_none = filter_none
         self.closing_sentinel = Closed
 
+    __summary_ids_mapping = WeakKeyDefaultDict(dict)
+    __id_gen = itertools.count()
+
+    def __get_summary_ids(self, widget: 'WidgetSignalsMixin'):
+        ids = self.__summary_ids_mapping[widget]
+        return ids.setdefault(self.name, [])
+
     def __call__(self, method):
         def summarize_wrapper(widget, index, value):
-            widget.set_partial_input_summary(
-                self.name, summarize(value), id=index)
+            # If this method is overridden, don't summarize
+            if summarize_wrapper is getattr(type(widget), method.__name__):
+                ids = self.__get_summary_ids(widget)
+                widget.set_partial_input_summary(
+                    self.name, summarize(value), id=ids[index], index=index)
             method(widget, index, value)
         _ = super().__call__(method)
         return summarize_wrapper if self.auto_summary else method
@@ -267,9 +281,11 @@ class MultiInput(Input):
     def insert(self, method):
         """Register the method as the insert handler"""
         def summarize_wrapper(widget, index, value):
-            self._insert_summary_slot(widget, index)
-            widget.set_partial_input_summary(
-                self.name, summarize(value), id=index)
+            if summarize_wrapper is getattr(type(widget), method.__name__):
+                ids = self.__get_summary_ids(widget)
+                ids.insert(index, next(self.__id_gen))
+                widget.set_partial_input_summary(
+                    self.name, summarize(value), id=ids[index], index=index)
             method(widget, index, value)
         self.insert_handler = method.__name__
         return summarize_wrapper if self.auto_summary else method
@@ -277,9 +293,11 @@ class MultiInput(Input):
     def remove(self, method):
         """"Register the method as the remove handler"""
         def summarize_wrapper(widget, index):
-            self._remove_summary_slot(widget, index)
-            widget.set_partial_input_summary(
-                self.name, summarize(None), id=id)
+            if summarize_wrapper is getattr(type(widget), method.__name__):
+                ids = self.__get_summary_ids(widget)
+                id_ = ids.pop(index)
+                widget.set_partial_input_summary(
+                    self.name, summarize(None), id=id_)
             method(widget, index)
         self.remove_handler = method.__name__
         return summarize_wrapper if self.auto_summary else method
@@ -291,20 +309,6 @@ class MultiInput(Input):
             raise RuntimeError('remove_handler is not set')
         return super().bound_signal(widget)
 
-    def _insert_summary_slot(self, widget, index):
-        summaries = widget.input_summaries[self.name]
-        values = list(summaries.values())
-        values.insert(index, None)
-        summaries = {i: v for i, v in enumerate(values)}  # renumerate
-        widget.input_summaries[self.name] = summaries
-
-    def _remove_summary_slot(self, widget, index):
-        summaries = widget.input_summaries[self.name]
-        values = list(summaries.values())
-        del values[index]
-        summaries = {i: v for i, v in enumerate(values)}  # renumerate
-        widget.input_summaries[self.name] = summaries
-
 
 _not_set = object()
 
@@ -515,21 +519,28 @@ class WidgetSignalsMixin:
         self._update_summary(self.input_summaries)
         self._update_summary(self.output_summaries)
 
-    def set_partial_input_summary(self, name, partial_summary, *, id=None):
-        self._set_part_summary(self.input_summaries[name], id, partial_summary)
+    def set_partial_input_summary(self, name, partial_summary, *, id=None, index=None):
+        self.__set_part_summary(self.input_summaries[name], id, partial_summary, index=index)
         self._update_summary(self.input_summaries)
 
     def set_partial_output_summary(self, name, partial_summary, *, id=None):
-        self._set_part_summary(self.output_summaries[name], id, partial_summary)
+        self.__set_part_summary(self.output_summaries[name], id, partial_summary)
         self._update_summary(self.output_summaries)
 
     @staticmethod
-    def _set_part_summary(summary, id, partial_summary):
+    def __set_part_summary(summary, id, partial_summary, index=None):
         if partial_summary.summary is None:
             if id in summary:
                 del summary[id]
         else:
-            summary[id] = partial_summary
+            if index is None or id in summary:
+                summary[id] = partial_summary
+            else:
+                # Insert inplace at specified index
+                items = list(summary.items())
+                items.insert(index, (id, partial_summary))
+                summary.clear()
+                summary.update(items)
 
     def _update_summary(self, summaries):
         from orangewidget.widget import StateInfo
@@ -653,7 +664,7 @@ def set_multi_input_helper(
     inout.closing_sentinel as the obj.
     """
     inputs_ = get_widget_inputs(widget)
-    inputs = list(inputs_.setdefault(input.name, ()))
+    inputs = inputs_.setdefault(input.name, ())
     filter_none = input.filter_none
 
     signal_old = None
@@ -668,15 +679,16 @@ def set_multi_input_helper(
         index = key_to_pos.get(key)
         assert index is not None
 
+    inputs_updated = list(inputs)
     if new:
-        inputs.insert(index, (key, obj))
+        inputs_updated.insert(index, (key, obj))
     elif remove:
-        signal_old = inputs.pop(index)
+        signal_old = inputs_updated.pop(index)
     else:
-        signal_old = inputs[index]
-        inputs[index] = (key, obj)
+        signal_old = inputs_updated[index]
+        inputs_updated[index] = (key, obj)
 
-    inputs_[input.name] = tuple(inputs)
+    inputs_[input.name] = tuple(inputs_updated)
 
     if filter_none:
         def filter_f(obj):
@@ -704,10 +716,16 @@ def set_multi_input_helper(
         if new and filtered:
             # insert in inputs only (done above)
             return
+        elif new:
+            # Some inputs before this might be filtered invalidating the
+            # effective index. Find appropriate index for insertion
+            index = len([obj for _, obj in inputs[:index] if not filter_f(obj)])
         elif remove:
             if filter_f(signal_old[1]):
-                # was already removed, only remove from inputs (done above)
+                # was already notified as removed, only remove from inputs (done above)
                 return
+            else:
+                index = local_index(key, inputs, filter_f)
         elif update and filtered:
             if filter_f(signal_old[1]):
                 # did not change; remains filtered
@@ -718,6 +736,8 @@ def set_multi_input_helper(
                 new = False
                 index = local_index(key, inputs, filter_f)
                 assert index is not None
+        elif update:
+            index = local_index(key, inputs, filter_f)
 
         if signal_old is not None and filter_f(signal_old[1]) and not filtered:
             # update with non-none value, substitute as new signal
